﻿/*
 * Copyright (c) 2023 Proton AG
 *
 * This file is part of ProtonVPN.
 *
 * ProtonVPN is free software:
 you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * ProtonVPN is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
 */

using System.Net;
using System.Net.Http;
using System.Security;
using System.Threading.Tasks;
using NSubstitute;
using ProtonVPN.Core.Auth;
using ProtonVPN.Core.Srp;
using RichardSzalay.MockHttp;

namespace ProtonVPN.IntegrationTests
{
    public class AuthenticatedUserTests : TestBase
    {
        protected const string USERNAME = "username";
        protected const string CORRECT_PASSWORD = "password";
        protected const string WRONG_PASSWORD = "wrong";
        protected const string CERTIFICATE =
            "-----BEGIN CERTIFICATE-----\n" +
            "MIICEzCCAcWgAwIBAgICA6QwBQYDK2VwMDgxNjA0BgNVBAMMLVByb3RvblZQTiBD\n" +
            "bGllbnQgQ2VydGlmaWNhdGUgQXV0aG9yaXR5IDE3ODYxMzAeFw0yMjA2MzAwOTIy\n" +
            "NDZaFw0yMjA3MDEwOTIyNDdaMA4xDDAKBgNVBAMMAzkzMjAqMAUGAytlcAMhAIRl\n" +
            "nVhh8k5SlmSWqEjvaLnghx2z01fD/jWtWikfyJ3oo4IBGzCCARcwHQYDVR0OBBYE\n" +
            "FD2ziAiZ58wvQokqBl/YrHX3rU9gMBMGDCsGAQQBg7tpAQAAAAQDAgEAMBMGDCsG\n" +
            "AQQBg7tpAQAAAQQDAgEAMBwGDCsGAQQBg7tpAQAAAgQMMAoECHZwbi1mcmVlMBkG\n" +
            "DCsGAQQBg7tpAQAAAwQJBAdXaW5kb3dzMA4GA1UdDwEB/wQEAwIHgDAMBgNVHRMB\n" +
            "Af8EAjAAMBMGA1UdJQQMMAoGCCsGAQUFBwMCMGAGA1UdIwRZMFeAFAM/RhKQDNy0\n" +
            "eNyeSWDEioh6Rv1goTykOjA4MTYwNAYDVQQDDC1Qcm90b25WUE4gQ2xpZW50IENl\n" +
            "cnRpZmljYXRlIEF1dGhvcml0eSAxNzg2MTOCAQIwBQYDK2VwA0EATuIkjDSzmZa5\n" +
            "UIfgGDGcNvFTVQBxSNzmGgZ+8YS60asMpo2vlGOyuy1U8vb3PsHSRYIgnJVup0P0\n" +
            "aFJlft9QCg==\n" +
            "-----END CERTIFICATE-----\n";
        protected const string SSO_RESPONSE_TOKEN = "wfih0367aa7dc0359bf5c42d15a93e6c";

        protected void SetApiResponsesForAuth()
        {
            SetUnauthSessionResponse();
            SetAuthInfoResponse();
            SetAuthResponse();
            SetUserResponse();
            SetVpnInfoResponse();
            SetAuthCertificateResponse();
            InitializeContainer();
        }

        protected void SetApiResponsesForSsoAuth()
        {
            SetUnauthSessionResponse();
            SetAuthInfoResponse();
            SetAuthResponse();
            SetUserResponse();
            SetVpnInfoResponse();
            InitializeContainer();

        }

        protected void SetApiResponsesForAuthWithTwoFactor()
        {
            SetUnauthSessionResponse();
            SetAuthInfoResponse();
            SetAuthResponseWithTwoFactorEnabled();
            SetUserResponse();
            InitializeContainer();
        }

        private void SetUnauthSessionResponse()
        {
            MessageHandler.When(HttpMethod.Post, "/auth/v4/sessions").Respond(_ => new HttpResponseMessage
            {
                StatusCode = HttpStatusCode.OK,
                Content = new StringContent(GetJsonMock("UnauthSessionResponseMock"))
            });
        }

        private void SetAuthInfoResponse()
        {
            MessageHandler.When(HttpMethod.Post, "/auth/info").Respond(_ => new HttpResponseMessage
            {
                StatusCode = HttpStatusCode.OK,
                Content = new StringContent(GetJsonMock("AuthInfoResponseMock"))
            });
        }

        private void SetAuthResponse()
        {
            MessageHandler.When(HttpMethod.Post, "/auth").Respond(_ => new HttpResponseMessage
            {
                StatusCode = HttpStatusCode.OK,
                Content = new StringContent(GetJsonMock("AuthResponseMock"))
            });
        }

        private void SetAuthResponseWithTwoFactorEnabled()
        {
            MessageHandler.When(HttpMethod.Post, "/auth").Respond(_ => new HttpResponseMessage
            {
                StatusCode = HttpStatusCode.OK,
                Content = new StringContent(GetJsonMock("AuthResponseWithTwoFactorEnabledMock"))
            });
        }

        private void SetVpnInfoResponse()
        {
            MessageHandler.When(HttpMethod.Get, "/vpn/v2").Respond(_ => new HttpResponseMessage
            {
                StatusCode = HttpStatusCode.OK,
                Content = new StringContent(GetJsonMock("VpnInfoWrapperResponseMock"))
            });
        }

        private void SetAuthCertificateResponse()
        {
            MessageHandler.When(HttpMethod.Post, "/vpn/v1/certificate").Respond(_ => new HttpResponseMessage
            {
                StatusCode = HttpStatusCode.OK,
                Content = new StringContent(GetJsonMock("CertificateResponseMock"))
            });
        }

        private void SetUserResponse()
        {
            MessageHandler.When(HttpMethod.Post, "/core/v4/users").Respond(_ => new HttpResponseMessage
            {
                StatusCode = HttpStatusCode.OK,
                Content = new StringContent(GetJsonMock("UserResponseMock"))
            });
        }

        protected async Task<AuthResult> MakeUserAuth(string username, string password)
        {
            SrpPInvoke.SetUnitTest();
            SecureString securePassword = new NetworkCredential(username, password).SecurePassword;
            return await Resolve<IUserAuthenticator>().LoginUserAsync(username, securePassword);
        }

        protected async Task<AuthResult> MakeUserAuthWithSso(string username)
        {
            Resolve<ISsoAuthenticator>().AuthenticateAsync(Arg.Any<AuthSsoSessionInfo>())
                .Returns(SSO_RESPONSE_TOKEN);

            return await Resolve<IUserAuthenticator>().SingleSignOnUserAsync(username);
        }
    }
}